@@ -1,3 +1,5 @@
+#include <isl/aff.h>
+
 #define xFN(TYPE,NAME) TYPE ## _ ## NAME
 #define FN(TYPE,NAME) xFN(TYPE,NAME)
 #define xS(TYPE,NAME) struct TYPE ## _ ## NAME
@@ -190,6 +192,11 @@ __isl_give isl_id *FN(PW,get_dim_id)(__isl_keep PW *pw, enum isl_dim_type type,
 	return pw ? isl_space_get_dim_id(pw->dim, type, pos) : NULL;
 }
 
+int FN(PW,has_tuple_name)(__isl_keep PW *pw, enum isl_dim_type type)
+{
+	return pw ? isl_space_has_tuple_name(pw->dim, type) : -1;
+}
+
 const char *FN(PW,get_tuple_name)(__isl_keep PW *pw, enum isl_dim_type type)
 {
 	return pw ? isl_space_get_tuple_name(pw->dim, type) : NULL;
@@ -221,7 +228,7 @@ __isl_give PW *FN(PW,realign_domain)(__isl_take PW *pw,
 
 	pw = FN(PW,cow)(pw);
 	if (!pw || !exp)
-		return NULL;
+		goto error;
 
 	for (i = 0; i < pw->n; ++i) {
 		pw->p[i].set = isl_set_realign(pw->p[i].set,
@@ -531,9 +538,9 @@ static __isl_give PW *FN(PW,on_shared_domain_in)(__isl_take PW *pw1,
 
 	n = pw1->n * pw2->n;
 #ifdef HAS_TYPE
-	res = FN(PW,alloc_size)(space, pw1->type, n);
+	res = FN(PW,alloc_size)(isl_space_copy(space), pw1->type, n);
 #else
-	res = FN(PW,alloc_size)(space, n);
+	res = FN(PW,alloc_size)(isl_space_copy(space), n);
 #endif
 
 	for (i = 0; i < pw1->n; ++i) {
@@ -561,6 +568,7 @@ static __isl_give PW *FN(PW,on_shared_domain_in)(__isl_take PW *pw1,
 		}
 	}
 
+	isl_space_free(space);
 	FN(PW,free)(pw1);
 	FN(PW,free)(pw2);
 	return res;
@@ -643,7 +651,8 @@ __isl_give isl_qpolynomial *FN(PW,eval)(__isl_take PW *pw,
 		goto error;
 	ctx = isl_point_get_ctx(pnt);
 	pnt_dim = isl_point_get_space(pnt);
-	isl_assert(ctx, isl_space_is_domain(pnt_dim, pw->dim), goto error);
+	isl_assert(ctx, isl_space_is_domain_internal(pnt_dim, pw->dim),
+		    goto error);
 
 	for (i = 0; i < pw->n; ++i) {
 		found = isl_set_contains_point(pw->p[i].set, pnt);
@@ -796,16 +805,17 @@ static __isl_give PW *FN(PW,gist_aligned)(__isl_take PW *pw,
 		goto error;
 
 	for (i = pw->n - 1; i >= 0; --i) {
-		pw->p[i].set = isl_set_intersect(pw->p[i].set,
+		isl_set *set_i;
+		int empty;
+
+		set_i = isl_set_intersect(isl_set_copy(pw->p[i].set),
 						 isl_set_copy(context));
-		if (!pw->p[i].set)
-			goto error;
-		pw->p[i].FIELD = fn_el(pw->p[i].FIELD,
-					     isl_set_copy(pw->p[i].set));
+		empty = isl_set_plain_is_empty(set_i);
+		pw->p[i].FIELD = fn_el(pw->p[i].FIELD, set_i);
 		pw->p[i].set = fn_dom(pw->p[i].set, isl_basic_set_copy(hull));
-		if (!pw->p[i].set)
+		if (!pw->p[i].FIELD || !pw->p[i].set)
 			goto error;
-		if (isl_set_plain_is_empty(pw->p[i].set)) {
+		if (empty) {
 			isl_set_free(pw->p[i].set);
 			FN(EL,free)(pw->p[i].FIELD);
 			if (i != pw->n - 1)
@@ -973,12 +983,14 @@ __isl_give PW *FN(PW,drop_dims)(__isl_take PW *pw,
 	if (!pw->dim)
 		goto error;
 	for (i = 0; i < pw->n; ++i) {
-		pw->p[i].set = isl_set_drop(pw->p[i].set, set_type, first, n);
-		if (!pw->p[i].set)
-			goto error;
 		pw->p[i].FIELD = FN(EL,drop_dims)(pw->p[i].FIELD, type, first, n);
 		if (!pw->p[i].FIELD)
 			goto error;
+		if (type == isl_dim_out)
+			continue;
+		pw->p[i].set = isl_set_drop(pw->p[i].set, set_type, first, n);
+		if (!pw->p[i].set)
+			goto error;
 	}
 
 	return pw;
@@ -1302,7 +1314,7 @@ __isl_give PW *FN(PW,morph_domain)(__isl_take PW *pw,
 		goto error;
 
 	ctx = isl_space_get_ctx(pw->dim);
-	isl_assert(ctx, isl_space_is_domain(morph->dom->dim, pw->dim),
+	isl_assert(ctx, isl_space_is_domain_internal(morph->dom->dim, pw->dim),
 		goto error);
 
 	pw = FN(PW,cow)(pw);
@@ -1519,8 +1531,8 @@ __isl_give PW *FN(PW,scale)(__isl_take PW *pw, isl_int v)
 
 static int FN(PW,qsort_set_cmp)(const void *p1, const void *p2)
 {
-	const isl_set *set1 = *(const isl_set **)p1;
-	const isl_set *set2 = *(const isl_set **)p2;
+	isl_set *set1 = *(isl_set * const *)p1;
+	isl_set *set2 = *(isl_set * const *)p2;
 
 	return isl_set_plain_cmp(set1, set2);
 }
@@ -1609,3 +1621,158 @@ error:
 	FN(PW,free)(pw2);
 	return -1;
 }
+
+#ifndef NO_PULLBACK
+static __isl_give PW *FN(PW,align_params_pw_multi_aff_and)(__isl_take PW *pw,
+	__isl_take isl_multi_aff *ma,
+	__isl_give PW *(*fn)(__isl_take PW *pw, __isl_take isl_multi_aff *ma))
+{
+	isl_ctx *ctx;
+	isl_space *ma_space;
+
+	ma_space = isl_multi_aff_get_space(ma);
+	if (!pw || !ma || !ma_space)
+		goto error;
+	if (isl_space_match(pw->dim, isl_dim_param, ma_space, isl_dim_param)) {
+		isl_space_free(ma_space);
+		return fn(pw, ma);
+	}
+	ctx = FN(PW,get_ctx)(pw);
+	if (!isl_space_has_named_params(pw->dim) ||
+	    !isl_space_has_named_params(ma_space))
+		isl_die(ctx, isl_error_invalid,
+			"unaligned unnamed parameters", goto error);
+	pw = FN(PW,align_params)(pw, ma_space);
+	ma = isl_multi_aff_align_params(ma, FN(PW,get_space)(pw));
+	return fn(pw, ma);
+error:
+	isl_space_free(ma_space);
+	FN(PW,free)(pw);
+	isl_multi_aff_free(ma);
+	return NULL;
+}
+
+static __isl_give PW *FN(PW,align_params_pw_pw_multi_aff_and)(__isl_take PW *pw,
+	__isl_take isl_pw_multi_aff *pma,
+	__isl_give PW *(*fn)(__isl_take PW *pw,
+		__isl_take isl_pw_multi_aff *ma))
+{
+	isl_ctx *ctx;
+	isl_space *pma_space;
+
+	pma_space = isl_pw_multi_aff_get_space(pma);
+	if (!pw || !pma || !pma_space)
+		goto error;
+	if (isl_space_match(pw->dim, isl_dim_param, pma_space, isl_dim_param)) {
+		isl_space_free(pma_space);
+		return fn(pw, pma);
+	}
+	ctx = FN(PW,get_ctx)(pw);
+	if (!isl_space_has_named_params(pw->dim) ||
+	    !isl_space_has_named_params(pma_space))
+		isl_die(ctx, isl_error_invalid,
+			"unaligned unnamed parameters", goto error);
+	pw = FN(PW,align_params)(pw, pma_space);
+	pma = isl_pw_multi_aff_align_params(pma, FN(PW,get_space)(pw));
+	return fn(pw, pma);
+error:
+	isl_space_free(pma_space);
+	FN(PW,free)(pw);
+	isl_pw_multi_aff_free(pma);
+	return NULL;
+}
+
+/* Compute the pullback of "pw" by the function represented by "ma".
+ * In other words, plug in "ma" in "pw".
+ */
+static __isl_give PW *FN(PW,pullback_multi_aff_aligned)(__isl_take PW *pw,
+	__isl_take isl_multi_aff *ma)
+{
+	int i;
+	isl_space *space = NULL;
+
+	ma = isl_multi_aff_align_divs(ma);
+	pw = FN(PW,cow)(pw);
+	if (!pw || !ma)
+		goto error;
+
+	space = isl_space_join(isl_multi_aff_get_space(ma),
+				FN(PW,get_space)(pw));
+
+	for (i = 0; i < pw->n; ++i) {
+		pw->p[i].set = isl_set_preimage_multi_aff(pw->p[i].set,
+						    isl_multi_aff_copy(ma));
+		if (!pw->p[i].set)
+			goto error;
+		pw->p[i].FIELD = FN(EL,pullback_multi_aff)(pw->p[i].FIELD,
+						    isl_multi_aff_copy(ma));
+		if (!pw->p[i].FIELD)
+			goto error;
+	}
+
+	pw = FN(PW,reset_space)(pw, space);
+	isl_multi_aff_free(ma);
+	return pw;
+error:
+	isl_space_free(space);
+	isl_multi_aff_free(ma);
+	FN(PW,free)(pw);
+	return NULL;
+}
+
+__isl_give PW *FN(PW,pullback_multi_aff)(__isl_take PW *pw,
+	__isl_take isl_multi_aff *ma)
+{
+	return FN(PW,align_params_pw_multi_aff_and)(pw, ma,
+					&FN(PW,pullback_multi_aff_aligned));
+}
+
+/* Compute the pullback of "pw" by the function represented by "pma".
+ * In other words, plug in "pma" in "pw".
+ */
+static __isl_give PW *FN(PW,pullback_pw_multi_aff_aligned)(__isl_take PW *pw,
+	__isl_take isl_pw_multi_aff *pma)
+{
+	int i;
+	PW *res;
+
+	if (!pma)
+		goto error;
+
+	if (pma->n == 0) {
+		isl_pw_multi_aff_free(pma);
+		res = FN(PW,empty)(FN(PW,get_space)(pw));
+		FN(PW,free)(pw);
+		return res;
+	}
+
+	res = FN(PW,pullback_multi_aff)(FN(PW,copy)(pw),
+					isl_multi_aff_copy(pma->p[0].maff));
+	res = FN(PW,intersect_domain)(res, isl_set_copy(pma->p[0].set));
+
+	for (i = 1; i < pma->n; ++i) {
+		PW *res_i;
+
+		res_i = FN(PW,pullback_multi_aff)(FN(PW,copy)(pw),
+					isl_multi_aff_copy(pma->p[i].maff));
+		res_i = FN(PW,intersect_domain)(res_i,
+					isl_set_copy(pma->p[i].set));
+		res = FN(PW,add_disjoint)(res, res_i);
+	}
+
+	isl_pw_multi_aff_free(pma);
+	FN(PW,free)(pw);
+	return res;
+error:
+	isl_pw_multi_aff_free(pma);
+	FN(PW,free)(pw);
+	return NULL;
+}
+
+__isl_give PW *FN(PW,pullback_pw_multi_aff)(__isl_take PW *pw,
+	__isl_take isl_pw_multi_aff *pma)
+{
+	return FN(PW,align_params_pw_pw_multi_aff_and)(pw, pma,
+					&FN(PW,pullback_pw_multi_aff_aligned));
+}
+#endif
